package neuroet;

import java.util.*;
import java.io.*;

public final class DataSet extends Object implements Serializable
{
	transient private DataElement[] originalData;  //has unique DataElements
	transient private DataElement[] scaledData;

	transient private DataElement[] testingData;  //uses arrays from scaledData
	transient private DataElement[] trainingData; //uses arrays from scaledData
	transient private DataElement[] validationData; //uses arrays from scaledData



	private Scaler[] inputScalers;
	private Scaler outputScaler;

        private String[] names;
        private String outputName;

	private final static long serialVersionUID= 556568756L;

	private int inputWidth;
	private int outputWidth;


        public String[] GetInputNames()
        {
          return (String[])names.clone();
        }

        public String GetOutputName()
        {
          return outputName;
        }

        public int GetInputWidth()
        {
          return inputWidth;
        }

        public int GetOutputWidth()
        {
          return outputWidth;
        }

        public String GetColumnName(int col)
        {
          return names[col];
        }

	DataSet (File input, File answers, PrefPanel p) throws IOException
	{
		this(input, answers, p.GetTestProportion(), p.GetValidationProportion(), p);

	}


	DataSet(File input, File answers, double testProp, double validProp, PrefPanel p) throws IOException
	{
		ReadData(input, answers);
		CreateScalers(p);
		ScaleData();
		CreateTestingAndTrainingData(testProp, validProp);
	}

        private DataSet(DataElement[] oData, Scaler[] iScalers, Scaler oScaler, String[] strs, String on)
        {
            originalData = oData;
            names = strs;
            outputName = on;


            inputWidth = originalData[0].inputs.length;
            outputWidth = 1;

            inputScalers = (Scaler[])iScalers.clone();

            if (inputScalers.length != inputWidth)
               throw new IllegalArgumentException("input length != Scalers length");

            if (strs.length != inputWidth)
              throw new IllegalArgumentException("input length != Names length");

            outputScaler = oScaler;
            ScaleData();
            CreateTestingAndTrainingData(0.0, 0.0);

        }

        public DataSet CloneDataSetMinusInput(int inp)
        {
          DataElement[] nDE = new DataElement[originalData.length];


          for (int x = 0; x < nDE.length; x++)
          {
            nDE[x] = (DataElement)originalData[x].clone();
            nDE[x].RemoveInput(inp);
          }

          Scaler[] nInputScalers = new Scaler[inputScalers.length - 1];

          System.arraycopy(inputScalers, 0, nInputScalers, 0, inp);
          System.arraycopy(inputScalers, inp + 1,nInputScalers, inp, nInputScalers.length - inp);

          String[] inpNames = new String[this.GetInputWidth() - 1];

          for (int x = 0, y = 0; x < inpNames.length; x++)
          {
            if (x != inp)
            {
                inpNames[y++] = names[x];//only increment y in this case
            }
          }

          return new DataSet(nDE, nInputScalers,outputScaler, inpNames, this.GetOutputName());
        }

        public DataSet ExtractOneInput(int inp)
        {
          return ExtractXInputs(new int[]{inp});
        }

        public DataSet ExtractXInputs(int[] inps)
        {
          Scaler[] inputScaler = new Scaler[inps.length];
          String[] inpNames = new String[inps.length];

          for (int x = 0; x < inputScaler.length; x++)
          {
            inputScaler[x] = inputScalers[inps[x]];
            inpNames[x] = names[inps[x]];
          }

          DataElement[] ar = new DataElement[originalData.length];

          for (int x = 0; x < ar.length; x++)
          {
            double[] dat = new double[inps.length];

            for (int y = 0; y < inps.length; y++)
            {
              dat[y] = originalData[x].GetInput(inps[y]);
            }

            ar[x] = new DataElement(dat, originalData[x].output);

          }

          return new DataSet(ar,inputScaler, outputScaler, inpNames, this.GetOutputName());


        }



	public void ReadData (File input, File answers) throws IOException
	{
          BufferedReader in = new BufferedReader(new FileReader(input));

          String f = null;

          for (f = in.readLine();; f = in.readLine())
          {
            if (f == null)
            {
                throw new IOException("Nothing found in input file");
            }

            f = f.trim();

            if (f.equals(""))
            {
              continue;
            }

            break;

          }

          names = Utilities.GetAllCategories(new StringTokenizer(f, "\t"));

          BufferedReader ans = new BufferedReader(new FileReader(answers));

          for (outputName = ans.readLine();; outputName = ans.readLine())
          {
            if (outputName == null)
            {
                throw new IOException("Nothing found in output file");
            }

            outputName = outputName.trim();

            if (outputName.equals(""))
            {
              continue;
            }

            break;

          }//reads in output name



                double[][] inputs = Utilities.ReadMatrix(Utilities.ReadAll(in));
                double[][] outputs = Utilities.ReadMatrix(Utilities.ReadAll(ans));

                ans.close();
                in.close();

                if (inputs.length != outputs.length)
                  throw new IOException("Different Number of Rows in input and output file");

                if (inputs.length == 0)
                  throw new IOException("Zero rows are not allowed");

                inputWidth = inputs[0].length;
                outputWidth = 1;

                if (inputWidth == 0)
                  throw new IOException("Width of Zero is not allowed");

                if (names.length != inputWidth)
                  throw new IOException("Different number of names to input: " + names.length + "\t" + inputWidth);

                for (int x = 0; x < inputs.length; x++)
                {
                  if (outputs[x].length != outputWidth)
                    throw new IOException("Output of length not equal to: " + outputWidth + " detected at row: " + (x + 1));

                  if (inputs[x].length != inputWidth)
                    throw new IOException("Input of length not equal to: " + inputWidth +  " detected at row: " + (x + 1));
                }




                originalData = new DataElement[inputs.length];

                for (int x = 0; x < originalData.length; x++)
                {
                  originalData[x] = new DataElement(inputs[x] ,outputs[x][0]);
                }

	}

	public void CreateTestingAndTrainingData(double propTest, double propValid)  //we copy the scaled data, shuffle it, and copy into testing and training
	{

		if (propTest < 0.0 || propValid < 0.0 || (propTest + propValid) > .75)
		{
			throw new IllegalArgumentException("Bad Arguments passed to CreateTestingAndTrainingData: " + propTest + "\t" + propValid);
		}

		DataElement[] temp = (DataElement[])scaledData.clone();
		testingData = new DataElement[(int)(propTest * scaledData.length)];
		validationData = new DataElement[(int)(propValid * scaledData.length)];

		trainingData = new DataElement[scaledData.length - testingData.length - validationData.length];


                Collections.shuffle(Arrays.asList(temp));

                /*for (int x = 0; x < temp.length; x++)  // shuffles the data
		{
			holder = temp[x];
			current = Math.abs(Utilities.rGen.nextInt() % temp.length);
			temp[x] = temp[current];
			temp[current] = holder;
		}*/


		System.arraycopy(temp, 0, testingData, 0, testingData.length);
		System.arraycopy(temp, testingData.length, validationData, 0, validationData.length);
		System.arraycopy(temp, testingData.length + validationData.length, trainingData, 0, trainingData.length);

	}


	public double[][] MakeAsScaledInputs(File input) throws IOException
	{
		//StringTokenizer st = new StringTokenizer(new String(Utilities.ReadWholeFile(input)));
		/******************
		*  new code here
		*  change made by Travis krick
		*  June 8, 2004 
		*  change so that category names are read in the input file
		*  change # 1
		**/
		BufferedReader in = new BufferedReader(new FileReader(input));

		String f = null;
	
		for (f = in.readLine();; f = in.readLine())
		{
			if (f == null)
			{
				throw new IOException("Nothing found in input file");
			}
	
			f = f.trim();
	
			if (f.equals(""))
			{
			  continue;
			}
	
			break;

		}
		Utilities.GetAllCategories(new StringTokenizer(f, "\t"));
		
		
		
		double[][] answer = Utilities.ReadMatrix(Utilities.ReadAll(in));
		// stops here  - change # 1	
		// instead of preexisting code: 
		// double[][] answer = Utilities.ReadMatrix(input);

                for (int x = 0; x < answer.length; x++)
                {
                  if (answer[x].length != inputWidth)
                    throw new IOException("Bad row length: " + answer[x].length);
                }



		for (int y = 0; y < answer.length; y++)
		{
			for (int x = 0; x < inputWidth; x++)
			{
                          answer[y][x] = inputScalers[x].ScaleSingle(answer[y][x]);
			}
		}

		return answer;

	}

	public double[] RevertOutputs(double[] scaledOutputs)
	{
		//if (scaledOutputs[0].length != outputWidth)
		//	throw new IllegalArgumentException("Width of 2-d array not equal to outputWidth: " + scaledOutputs[0].length + "\t" + outputWidth);


		return outputScaler.RevertArray(scaledOutputs);


	}

	private void CreateScalers(PrefPanel p)
	{

		inputScalers = p.GetInputScalers(ArrayUtilities.RotateMatrix(GetOriginalInputs()));

		outputScaler = p.GetOutputScaler(GetOriginalOutput());

		/*inputScalers = new Scaler[inputWidth];  //number of inputs plus the answer
		//outputScaler = new Scaler[outputWidth];



		for (int x = 0; x < inputWidth; x++)
		{
			inputScalers[x] = Scaler.GetDoNothingScaler(); //GetOriginalInputColumn(x)

	    }

		outputScaler = Scaler.GetDoNothingScaler();
		*/
	}

	public double[][] GetTrainingInputs()
	{
		double[][] answer = new double[trainingData.length][];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = trainingData[x].inputs;
		}
		return answer;
	}

	public double[] GetTrainingOutput()
	{
		double[] answer = new double[trainingData.length];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = trainingData[x].output;
		}

		return answer;
	}


        public double[] GetTrainingOutputUnscaled()
        {
          return RevertOutputs(GetTrainingOutput());
        }

	public double[][] GetValidationInputs()
	{
		double[][] answer = new double[validationData.length][];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = validationData[x].inputs;
		}


		return answer;
	}

	public double[] GetValidationOutput()
	{
		double[] answer = new double[validationData.length];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = validationData[x].output;
		}


		return answer;
	}

        public double[] GetValidationOutputUnscaled()
        {
          return RevertOutputs(GetValidationOutput());
        }

	public double[][] GetTestingInputs()
	{
		double[][] answer = new double[testingData.length][];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = testingData[x].inputs;
		}
		return answer;
	}

	public double[] GetTestingOutput()
	{
		double[] answer = new double[testingData.length];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = testingData[x].output;
		}

		return answer;
	}




        public double[] GetTestingOutputUnscaled()
        {
          return RevertOutputs(GetTestingOutput());
        }

	public double[][] GetScaledInputs()
	{
		double[][] answer = new double[scaledData.length][];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = scaledData[x].inputs;
		}

		return answer;
	}

	public double[] GetScaledOutput()
	{
		double[] answer = new double[scaledData.length];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = scaledData[x].output;
		}

		return answer;


	}

	public double[] GetOriginalOutput()
	{
		double[] answer = new double[originalData.length];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = originalData[x].output;
		}

		return answer;
	}

	public double[][] GetOriginalInputs()
	{

		double[][] answer = new double[originalData.length][];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = originalData[x].inputs;
		}

		return answer;

	}

	public double[] GetOriginalInputColumn (int column)
	{
		double[] answer = new double[originalData.length];

		for (int x = 0; x < answer.length; x++)
		{
			answer[x] = originalData[x].inputs[column];
		}

		return answer;

	}

	private void ScaleData()
	{
		scaledData = new DataElement[originalData.length];

		double[] inps;
		double outps;


		for (int x = 0; x < scaledData.length; x++)
		{
			inps = new double[inputWidth];
			//outps = new double[outputWidth];

			for (int y = 0; y < inputWidth; y++) //scale the inputs of an entry
			{
				inps[y] = inputScalers[y].ScaleSingle(originalData[x].inputs[y]);
			}

			outps = outputScaler.ScaleSingle(originalData[x].output);

			/*
			for (int y = 0; y < outputWidth; y++) //scale the outputs of an entry
			{
				outps[y] = outputScaler.ScaleSingle(originalData[x].output);
			}*/

			scaledData[x] = new DataElement(inps, outps);  //insert our new entry

		}



	}

        public String GetOriginalDataAsString()
        {
          return GetDataAsString(originalData);
        }

        public String GetScaledDataAsString()
        {
          return GetDataAsString(scaledData);
        }

        public String GetTrainingDataAsString()
        {
          return GetDataAsString(trainingData);
        }

        public String GetTestingDataAsString()
        {
          return GetDataAsString(testingData);
        }

        public String GetValidationDataAsString()
        {
          return GetDataAsString(validationData);
        }

        static private String GetDataAsString(DataElement[] ar)
        {
          StringWriter sw = new StringWriter();

          PrintWriter out = new PrintWriter(sw);

          for (int x = 0; x < ar.length; x++)
          {
            out.println(ar[x]);
          }

          out.close();

          return sw.getBuffer().toString();

        }
        public static void PrintDataSetToFiles(DataSet data, File dOut, File d2x, File d2y) throws IOException
        {

            PrintWriter dataOut = new PrintWriter(new FileWriter(dOut));
            PrintWriter Out2X = new PrintWriter(new FileWriter(d2x));
            PrintWriter Out2Y = new PrintWriter(new FileWriter(d2y));

            try{




            final int limit = data.GetInputWidth();

                        for (int x = 0; x < limit; x++)
                        {
                            dataOut.print (data.GetColumnName(x) + "\t");
                        }

                        dataOut.println(data.GetOutputName());


                        dataOut.println();
                        dataOut.println();

                        dataOut.println("Raw Data:");

                        dataOut.println(data.GetOriginalDataAsString());
                        dataOut.println();

                        dataOut.println("Scaled Data:");

                        dataOut.println(data.GetScaledDataAsString());
                        dataOut.println();

                        dataOut.println("Training Data:");

                        dataOut.println(data.GetTrainingDataAsString());
                        dataOut.println();

                        dataOut.println("Testing Data:");

                        dataOut.println(data.GetTestingDataAsString());
                        dataOut.println();

                        dataOut.println("Validation Data:");
                        dataOut.println(data.GetValidationDataAsString());


                        /*dataOut.println();
                        dataOut.println("Correlation Analysis");

                        String[] toPrint = data.ExamineCorrelationsAsStrings();

                        for (int x = 0; x < toPrint.length; x++)
                        {
                         dataOut.println(toPrint[x]);
                        }*/

                        Out2X.println(data.GetInputWidth());
                        Utilities.PrintDoubleMatrix(Out2X, data.GetScaledInputs());//used to be get original


                        Out2Y.println(data.GetOutputWidth());
                        Utilities.PrintVerticalDoubleArray(Out2Y,data.GetScaledOutput());
            }finally{dataOut.close(); Out2X.close(); Out2Y.close();}
        }

        public String[] ExamineCorrelationsAsStrings()
        {
          String[] ans = new String[(inputWidth * (inputWidth - 1)) / 2]; //should always be an even / 2
          int z = 0;

          for (int y = 0; y < inputWidth - 1; y++)
          {
            for (int x = y + 1; x < inputWidth; x++)
            {
              StringBuffer sb = new StringBuffer("Input" + y + "&" + "Input" + x + "\t");
              try{
              sb.append(ArrayUtilities.CalculateRunCorrelation(GetOriginalInputColumn(y), GetOriginalInputColumn(x)));
              }catch (IllegalArgumentException iae){sb.append("Unknown: " + iae);}
              ans[z] = sb.toString();
              z++;
            }
          }

          return ans;
        }

}


final class DataElement extends Object implements Cloneable //used to implement serializable
{
        double[] inputs;
	final double output;

	DataElement(double[] i, double o)
	{
		inputs = (double[])i.clone();
		output = o;
	}


        public double GetInput(int i)
        {
          return inputs[i];
        }

        public Object clone()
        {
          return new DataElement(inputs, output);
        }

        public String toString()
        {
            StringBuffer toReturn = new StringBuffer(100);

            for (int x = 0; x < inputs.length; x++)
            {
                toReturn.append (inputs[x]);
                toReturn.append ("\t");
            }

           	toReturn.append(output);



            return (toReturn.toString());
        }

        void RemoveInput(int inp)//package private
        {
          double[] replacement = new double[inputs.length - 1];


          System.arraycopy(inputs, 0, replacement, 0, inp);
          System.arraycopy(inputs, inp + 1,replacement, inp, replacement.length - inp);


           inputs = replacement;
        }
}

